@@ -10,15 +10,18 @@ import itertools
 import os
 import sys
 import glob
+from copy import copy
 
 from bokeh.plotting import figure, show, save, output_file
-from bokeh.models import ColumnDataSource, Plot, LinearAxis, Grid, CrosshairTool, HoverTool, CustomJS, Div
+from bokeh.models import ColumnDataSource, Plot, LinearAxis, Grid, CrosshairTool, HoverTool, CustomJS, Div, Select
 from bokeh.models.glyphs import VArea, Line
 from bokeh.layouts import gridplot, row, column, layout, Spacer
 from bokeh.models import Legend, Span, Label, BoxAnnotation
 from bokeh.colors.named import red, green, blue, orange, lightskyblue, mediumpurple, mediumspringgreen, grey
 from bokeh.events import Tap, DoubleTap, ButtonClick
 
+from .flocklab import FlocklabError
+
 
 ###############################################################################
 
@@ -79,13 +82,11 @@ def trace2series(t, v):
         if (vNew[i] == 0 and vNew[i+1] == 0):
             tNewNew.append(tNew[i])
             vNewNew.append(np.nan)
+    tNewNew.append(tNew[-1])
+    vNewNew.append(vNew[-1])
     tNewNew = np.asarray(tNewNew)
     vNewNew = np.asarray(vNewNew)
 
-    # add 0 at the end of each trace (to make sure that hover is available on last edge of varea of real signal edge)
-    tNewNew = np.append(tNewNew, tNewNew[-1])
-    vNewNew = np.append(vNewNew, 0)
-
     return (tNewNew, vNewNew)
 
 def plotObserverGpio(nodeId, nodeData, pOld):
@@ -96,10 +97,11 @@ def plotObserverGpio(nodeId, nodeData, pOld):
         # plot_width=1200,
         plot_height=900,
         min_border=0,
-        tools=['xpan', 'xwheel_zoom', 'xbox_zoom', 'hover', 'save', 'reset'],
+        tools=['xpan', 'xwheel_zoom', 'xbox_zoom', 'hover', 'reset'],
         active_drag='xbox_zoom', # not working due to bokeh bug https://github.com/bokeh/bokeh/issues/8766
         active_scroll='xwheel_zoom',
         sizing_mode='stretch_both', # full screen
+        # output_backend="webgl",
     )
     length = len(nodeData)
     vareas = []
@@ -108,7 +110,8 @@ def plotObserverGpio(nodeId, nodeData, pOld):
             continue
         signalName = '{} (Node {})'.format(pin, nodeId)
         t, v, = trace2series(pinData['t'], pinData['v'])
-        source = ColumnDataSource(dict(x=t, y1=np.zeros_like(v)+length-i, y2=v+length-i))
+        t_abs, v_dummy, = trace2series(pinData['tAbs'], pinData['v'])
+        source = ColumnDataSource(dict(x=t, t_abs=t_abs, y1=np.zeros_like(v)+length-i, y2=v+length-i))
         # plot areas
         vareaGlyph = VArea(x="x", y1="y1", y2="y2", fill_color=colorMapping(pin))
         varea = p.add_glyph(source, vareaGlyph, name=signalName)
@@ -123,7 +126,11 @@ def plotObserverGpio(nodeId, nodeData, pOld):
 
 
     hover = p.select(dict(type=HoverTool))
-    hover.tooltips = OrderedDict([('Time', '@x{0.0000000} s'),('Signal','$name')])
+    hover.tooltips = OrderedDict([
+        ('Time (rel)', '@x{0.0000000} s'),
+        ('Time (abs)', '@t_abs{0.0000000} s'),
+        ('Signal','$name')
+    ])
 
     p.xgrid.grid_line_color = None
     p.ygrid.grid_line_color = None
@@ -156,10 +163,11 @@ def plotObserverPower(nodeId, nodeData, pOld):
         # plot_width=1200,
         plot_height=900,
         min_border=0,
-        tools=['xpan', 'xwheel_zoom', 'xbox_zoom', 'hover', 'save', 'reset'],
+        tools=['xpan', 'xwheel_zoom', 'xbox_zoom', 'hover', 'reset'],
         active_drag='xbox_zoom', # not working due to bokeh bug https://github.com/bokeh/bokeh/issues/8766
         active_scroll='xwheel_zoom',
         sizing_mode='stretch_both', # full screen
+        # output_backend="webgl",
     )
     source = ColumnDataSource(dict(
       t=nodeData['t'],
@@ -175,7 +183,7 @@ def plotObserverPower(nodeId, nodeData, pOld):
     p.add_glyph(source, line_p, name='{}'.format(nodeId))
     hover = p.select(dict(type=HoverTool))
     hover.tooltips = OrderedDict([
-      ('Time', '@t{0.00000000} s'),
+      ('Time (rel)', '@t{0.00000000} s'),
       ('V', '@v{0.000000} V'),
       ('I', '@i{0.000000} mA'),
       ('Power', '@p{0.000000} mW'),
@@ -332,11 +340,12 @@ def plotAll(gpioData, powerData, testNum, interactive=False):
         x_range=lastPlot.x_range,
         plot_height=0,
         min_border=0,
-        tools=['xpan', 'xwheel_zoom', 'xbox_zoom', 'hover', 'save', 'reset'],
+        tools=['xpan', 'xwheel_zoom', 'xbox_zoom', 'hover', 'reset'],
         active_drag='xbox_zoom', # not working due to bokeh bug https://github.com/bokeh/bokeh/issues/8766
         active_scroll='xwheel_zoom',
         height_policy='fit',
         width_policy='fit',
+        # output_backend="webgl",
     )
     source = ColumnDataSource(dict(x=[0, maxT], y1=[0, 0], y2=[0, 0]))
     vareaGlyph = VArea(x="x", y1="y1", y2="y2", fill_color='grey')
@@ -405,7 +414,7 @@ def plotAll(gpioData, powerData, testNum, interactive=False):
     z-index: 1;
     top: 125%;
     left: 50%;
-    margin-left: -110px;
+    margin-left: -150px;
     opacity: 0;
     transition: opacity 0.3s;
   }
@@ -453,10 +462,24 @@ def plotAll(gpioData, powerData, testNum, interactive=False):
         width_policy='fit',
         align='center'
     )
-    spaceDiv = Div(
+    spaceDiv1 = Div(
         text='<div width="30px"></div>',
         style={'background-color': 'yellow'},
-        width=60,
+        width=50,
+        width_policy='fixed',
+        height_policy='fit',
+    )
+    spaceDiv2 = Div(
+        text='<div width="30px"></div>',
+        style={'background-color': 'yellow'},
+        width=30,
+        width_policy='fixed',
+        height_policy='fit',
+    )
+    spaceDiv3 = Div(
+        text='<div width="30px"></div>',
+        style={'background-color': 'yellow'},
+        width=30,
         width_policy='fixed',
         height_policy='fit',
     )
@@ -504,18 +527,62 @@ def plotAll(gpioData, powerData, testNum, interactive=False):
         text='''<table>
           <tr>
             <td align="left">
-              <div class="tooltip" style="vetical-align:middle; cursor:default; color:grey">&#9432;<span class="tooltiptext"><b>Set marker</b>:<br/>enable &#9312; or &#9313; + click into plot<br/><b>Remove markers</b>:<br/>disable &#9312; and &#9313; + click into plot<br/><b>Reset plot</b>:<br/>double-click inside plot</span></div>
+              <div class="tooltip" style="vertical-align:middle; cursor:default; color:grey; font-size: 20px; padding: 2px;">&#9432;<span class="tooltiptext"><b>Set marker</b>:<br/>enable &#9312; or &#9313; + click into plot<br/><b>Remove markers</b>:<br/>disable &#9312; and &#9313; + click into plot<br/><b>Reset plot</b>:<br/>double-click inside plot</span></div>
           </tr>
         </table>{}'''.format(tooltipStyle),
         width=30,
         width_policy='fixed',
         height_policy='fit',
     )
+    zoomLut = OrderedDict([
+        ('Quick Zoom', None),
+        ('fit trace', None),
+        ('1h', 3600),
+        ('10m', 600),
+        ('1m', 60),
+        ('10s', 10),
+        ('1s', 1),
+        ('100ms', 100e-3),
+        ('10ms', 10e-3),
+        ('1ms', 1e-3),
+        ('100us', 100e-6),
+        ('10us', 10e-6),
+        ('1us', 1e-6),
+    ])
+    zoomSelect = Select(
+        # title='Zoom ',
+        options=list(zoomLut.keys()),
+        width=120,
+        width_policy='fixed',
+        height_policy='fit',
+    )
+    zoomSelectCallback = CustomJS(
+        args={
+            'zoomLut': zoomLut,
+            'resetStr': list(zoomLut.keys())[0],
+            'p': pTime,
+            'zoomSelect': zoomSelect,
+        },
+        code="""
+        // console.log(cb_obj.value);
+        if (cb_obj.value == resetStr) {
+            // do nothing
+        } else if (cb_obj.value == "fit trace") {
+            p.reset.emit();
+        } else {
+            var start = p.x_range.getv('start');
+            var end = start + zoomLut[cb_obj.value];
+            p.x_range.setv({"start": start, "end": end});
+        }
+        // reset drop-down
+        zoomSelect.value = resetStr;
+        """,
+    )
+    zoomSelect.js_on_change('value', zoomSelectCallback)
     titleLine = row(
-        [titleDiv, spaceDiv, measureDiv, infoDiv],
+        [titleDiv, spaceDiv1, infoDiv, spaceDiv2, measureDiv, spaceDiv3, zoomSelect],
         sizing_mode='scale_width',
         align='start',
-        # align='center',
     )
     # put together final layout of page
     finalLayout = column(
@@ -532,7 +599,7 @@ def plotAll(gpioData, powerData, testNum, interactive=False):
 
 
 
-def visualizeFlocklabTrace(resultPath, outputDir=None, interactive=False):
+def visualizeFlocklabTrace(resultPath, outputDir=None, interactive=False, showPps=False, showRst=False):
     '''Plots FlockLab results using bokeh.
     Args:
         resultPath: path to the flocklab results (unzipped)
@@ -557,8 +624,8 @@ def visualizeFlocklabTrace(resultPath, outputDir=None, interactive=False):
     gpioAvailable = False
 
     if os.path.isfile(gpioPath):
-        # Read gpio data csv to pandas dataframe
-        gpioDf = pd.read_csv(gpioPath)
+        # Read gpio data csv to pandas dataframe (instruct pandas with float_precision to not sacrifice accuracy for the sake of speed)
+        gpioDf = pd.read_csv(gpioPath, float_precision='round_trip')
         # sanity check: column names
         for col in requiredGpioCols:
             if not col in gpioDf.columns:
@@ -580,8 +647,8 @@ def visualizeFlocklabTrace(resultPath, outputDir=None, interactive=False):
     powerAvailable = False
 
     if os.path.isfile(powerPath):
-        # Read power data csv to pandas dataframe
-        powerDf = pd.read_csv(powerPath)
+        # Read power data csv to pandas dataframe (instruct pandas with float_precision to not sacrifice accuracy for the sake of speed)
+        powerDf = pd.read_csv(powerPath, float_precision='round_trip')
         # sanity check: column names
         for col in requiredPowerCols:
             if not col in powerDf.columns:
@@ -618,7 +685,8 @@ def visualizeFlocklabTrace(resultPath, outputDir=None, interactive=False):
         if len(powerDf) > 0:
             powerAvailable = True
     else:
-        print('powerprofiling data (\'.csv\' or \'.rld\') could not be found!')
+        # print('powerprofiling data (\'.csv\' or \'.rld\') could not be found!')
+        pass
 
     # handle case where there is no data to plot
     if (not gpioAvailable) and (not powerAvailable):
@@ -636,31 +704,41 @@ def visualizeFlocklabTrace(resultPath, outputDir=None, interactive=False):
 
     # prepare gpio data
     gpioData = OrderedDict()
-    pinOrdering = ['INT1', 'INT2', 'LED1', 'LED2', 'LED3', 'SIG1', 'SIG2']
+    pinOrdering = ['INT1', 'INT2', 'LED1', 'LED2', 'LED3', 'SIG1', 'SIG2', 'PPS', 'nRST']
     if gpioAvailable:
         gpioDf['timestampRelative'] = gpioDf.timestamp - minT
         gpioDf.sort_values(by=['node_id', 'pin_name', 'timestamp'], inplace=True)
-
-        # Get overview of available data
-        gpioNodeList = sorted(list(set(gpioDf.node_id)))
-        gpioPinList = sorted(list(set(gpioDf.pin_name)))
-        # print('gpioNodeList:', gpioNodeList)
-        # print('gpioPinList:', gpioPinList)
+        # determine global end of gpio trace (for adding edge back to 0 at the end of trace for signals which end with 1)
+        tEnd = gpioDf[(gpioDf.pin_name=='nRST') & (gpioDf.value==0)].timestampRelative.to_numpy()[-1]
+        tAbsEnd = gpioDf[(gpioDf.pin_name=='nRST') & (gpioDf.value==0)].timestamp.to_numpy()[-1]
 
         # Generate gpioData dict from pandas dataframe
         for nodeId, nodeGrp in gpioDf.groupby('node_id'):
             # print(nodeId)
+            pinList = copy(pinOrdering)
             nodeData = OrderedDict()
             pinGrps = nodeGrp.groupby('pin_name')
-            pinList = pinOrdering + list(set(pinGrps.groups.keys()) - set(pinOrdering))
-            # TODO: test with gpio tracing file with missin  pin
+            if not set(pinGrps.groups.keys()).issubset(set(pinOrdering)):
+                raise FlocklabError('ERROR: GPIO tracing file contains unknown pin names!')
+            if not showRst:
+                pinList.remove('nRST')
+            if not showPps:
+                pinList.remove('PPS')
             for pin in pinList:
                 if pin in pinGrps.groups.keys():
                     pinGrp = pinGrps.get_group(pin)
-                    trace = {'t': pinGrp.timestampRelative.to_numpy(), 'v': pinGrp.value.to_numpy()}
-                    nodeData.update({pin: trace})
-                else:
-                    nodeData.update({pin: {}})
+                    # check if pin is ever toggled to 1 in the whole GPIO tracing (all nodes)
+                    if (gpioDf[(gpioDf.pin_name==pin)].value==1).any():
+                        t = pinGrp.timestampRelative.to_numpy()
+                        tAbs = pinGrp.timestamp.to_numpy()
+                        v = pinGrp.value.to_numpy()
+                        if len(v):
+                            if v[-1] == 1:
+                                t = np.append(t, tEnd)
+                                tAbs = np.append(tAbs, tAbsEnd)
+                                v = np.append(v, 0)
+                        trace = {'t': t, 'tAbs': tAbs, 'v': v}
+                        nodeData.update({pin: trace})
             gpioData.update({nodeId: nodeData})
 
 
